package pl.mpak.sky.gui.swing.syntax.structure;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import pl.mpak.sky.gui.swing.syntax.SyntaxDocument;
import pl.mpak.sky.gui.swing.syntax.SyntaxEditor.TokenRef;
import pl.mpak.sky.gui.swing.syntax.structure.core.ParserException;
import pl.mpak.util.Assert;

public abstract class StructureParser {

  private List<TokenRef> list;
  protected Iterator<TokenRef> iterator;
  protected TokenRef currentToken;
  protected TokenRef lastToken;
  protected boolean ignoreCase;
  protected HashMap<String, Integer> keyWordList;
  
  public StructureParser() {
    this.keyWordList = new HashMap<String, Integer>(); 
  }
  
  public void setTokenList(List<TokenRef> list) {
    if (this.list != null) {
      this.list.clear();
    }
    this.list = list;
    this.iterator = list.iterator();
    this.keyWordList.clear();
    this.currentToken = null;
    this.lastToken = null;
  }
  
  public void setIgnoreCase(boolean ignoreCase) {
    this.ignoreCase = ignoreCase;
  }
  
  /**
   * <p>Zwraca nastpny element lub null jeli koniec
   * @return
   * @throws ParserException 
   * @see getToken().prev
   * @see getToken().next
   * @see getToken()
   */
  protected TokenRef nextToken() throws ParserException {
    lastToken = currentToken;
    if (iterator.hasNext()) {
      return currentToken = iterator.next();
    }
    return currentToken = null;
  }
  
  /**
   * <p>Zwraca nastpny element i przy okazji pomija wszystko co zostao zdefiniowane w skipBlank()
   * @param skipBlanks
   * @return
   * @throws ParserException
   * @see skipBlank()
   */
  protected TokenRef nextToken(boolean skipBlanks) throws ParserException {
    TokenRef result = nextToken();
    if (skipBlanks && result != null) {
      result = skipBlank();
    }
    return result;
  }
  
  /**
   * <p>Domylnie pomija elementy SyntaxDocument.NONE ale mona funkcj przej i pomija inne elementy
   * @return
   * @throws ParserException
   * @see nextToken(boolean skipBlanks)
   */
  protected TokenRef skipBlank() throws ParserException {
    if (getToken().styleId == SyntaxDocument.NONE) {
      while (nextToken() != null) {
        if (getToken().styleId != SyntaxDocument.NONE) {
          return getToken();
        }
      }
    }
    return currentToken;
  }
  
  /**
   * <p>Zwraca poprzednio uzyskany token.<br>
   * Nie myli z getToken().prev
   * @return
   * @throws ParserException
   */
  protected TokenRef getLastToken() throws ParserException {
    if (lastToken == null) {
      throw new ParserException("Nieoczekiwane zakoczenie kodu");
    }
    return lastToken;
  }
  
  /**
   * <p>Zwraca offset dla okrelenia pocztku bloku
   * @return
   * @throws ParserException
   */
  protected int getStartOffset() throws ParserException {
    return getToken().offset;
  }
  
  /**
   * <p>Zwraca offset dla okrelenia koca bloku
   * @return
   * @throws ParserException
   */
  protected int getEndOffset() throws ParserException {
    return getLastToken().offset +getLastToken().token.length();
  }
  
  /**
   * <p>Sprawdzenie czy token jest identyczny co text
   * @param text
   * @return
   * @throws ParserException
   * @see Parser(Iterator&lt;TokenRef&gt; iterator, boolean <b>ignoreCase</b>)
   */
  protected boolean isToken(String text) throws ParserException {
    if (ignoreCase) {
      return getToken().isTokenIgnoreCase(text);
    }
    return getToken().isToken(text);
  }
  
  /**
   * <p>Zwraca cig znakw aktualnego tokenu
   * @return
   * @throws ParserException
   */
  protected String getTokenString() throws ParserException {
    return getToken().token;
  }
  
  /**
   * <p>Pozwla zmieni styleId elementu zalenego (w edytorze)
   * @param styleId
   * @throws ParserException
   */
  protected void setStyle(int styleId) throws ParserException {
    getToken().ref.styleId = styleId;
  }
  
  /**
   * <p>Zwraca bierzcy token
   * @return
   * @throws ParserException
   */
  protected TokenRef getToken() throws ParserException {
    if (currentToken == null) {
      throw new ParserException("Nieoczekiwane zakoczenie kodu");
    }
    return currentToken;
  }
  
  /**
   * <p>Umieszcza sowo na specjalnej licie, ktra na kocu parsowania moe by uyta do ustawienia styli tokenw
   * @param keyWord
   * @param styleId
   * @throws ParserException
   * @see updateKeyWordStyles
   */
  protected void putKeyWord(String keyWord, int styleId) throws ParserException {
    keyWordList.put((ignoreCase ? keyWord.toUpperCase() : keyWord), styleId);
    if (isToken(keyWord)) {
      setStyle(styleId);
    }
  }
  
  protected boolean updateStyle(int orygStyleId) {
    return orygStyleId == SyntaxDocument.NONE;
  }
  
  /**
   * <p>Aktualizuje tokeny edytora stylami okrelonymi na licie sw kluczowych
   * @throws ParserException
   * @see putKeyWord
   */
  protected void updateKeyWordStyles() {
    iterator = list.iterator();
    try {
      while (nextToken() != null) {
        Integer styleId = keyWordList.get((ignoreCase ? currentToken.token.toUpperCase() : currentToken.token));
        if (styleId != null && updateStyle(getToken().styleId)) {
          setStyle(styleId);
        }
      }
    } catch (ParserException e) {
    }
  }

  /**
   * <p>Gwna funkcja do parsowania
   * @return
   * @throws ParserException
   */
  public abstract BlockElement parse() throws ParserException;
  
}
